// SparseApproximateInverse.h
// Copyright (C) 2010 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _SparseApproximateInverse_h_
#define _SparseApproximateInverse_h_

#include <Basic/Log.h>
#include <Basic/Memory.h>
#include <Container/CSRMatrix.h>
#include <Container/CSVector.h>
#include <Container/List.h>
#include <Container/Matrix.h>
#include <Container/Vector.h>

#include <math.h>
#include <omp.h>


// E. Chow
// Parallel implementation and practical use of sparse approximate inverse preconditioners with a priori sparsity patterns
// International Journal of High Performance Computing, Vol 15. pp 56-74, 2001.

template <typename T>
void PowerPreconditionerPattern(const CSRMatrix<T>& A, CSRMatrix<T>& M, T threshold, int preconditioner_level, bool factorized, int threads) throw(Memory::Exception)
{
	try
	{
		Log(1, "PowerPreconditionerPattern, threshold = %e, level = %i, factorized = %i", threshold, preconditioner_level, factorized ? 1 : 0);
		// A must have a symetric structure

		int N = A.rows;

		// Scalling diagonal
		Vector<T> Ds(N);
		for (int i = 1; i <= N; ++i)
		{
			T v = fabs(A(i, i));
			Ds.entry[i] = v > 0 ? 1/sqrt(v) : 1;
		}

		// Create thrA (threshoding A)
		int nnz_thrA = 0;
		Vector<CSVector<int> > thrA(N);
		#pragma omp parallel for default(shared) reduction(+:nnz_thrA) schedule(guided) num_threads(threads)
		for (int i = 1; i <= N; ++i)
		{
			List<int> connectivity;
			int A_count_i = A.count[i];
			for (int c = 1; c <= A_count_i; ++c)
			{
				int j = A.index[i][c];
				if (i != j)
				{
					T v = fabs(Ds.entry[i]*A.entry[i][c]*Ds.entry[j]);
					if (v > threshold)
					{
						connectivity.AppendLast(j);
					}
				}
			}

			thrA.entry[i].Resize(N);
			thrA.entry[i].Allocate(connectivity.size);
			int c = 0;
			for (register ListItem<int>* conection = connectivity.first; conection; conection = conection->next)
			{
				++c;
				register int j = conection->value;
				thrA.entry[i].entry[c] = j;
			}
			nnz_thrA += connectivity.size;
		}
		Log(1, "Matrix threshoding, nnz(thrA) = %i", nnz_thrA);

		// Create pattern of powers sparse matrix
		Vector<Vector<int> > thread_index(threads);
		Vector<Vector<bool> > thread_used(threads);
		for (int t = 1; t <= threads; ++t)
		{
			thread_index.entry[t].Resize(N);
			thread_used.entry[t].Resize(N);
			thread_used.entry[t].Fill(false);
		}
		#pragma omp parallel for default(shared) schedule(guided) num_threads(threads)
		for (int i = 1; i <= N; ++i)
		{
			int thread = omp_get_thread_num() + 1;
			Vector<int>& index = thread_index.entry[thread];
			Vector<bool>& used = thread_used.entry[thread];

			int index_count = thrA.entry[i].count;
			for (int b = 1; b <= index_count; ++b)
			{
				int j = thrA.entry[i].entry[b];
				index.entry[b] = j;
				used.entry[j] = true;
			}
			used.entry[i] = true;

			int new_index_count = index_count;
			for (int l = 1; l <= preconditioner_level; ++l)
			{
				for (int b = 1; b <= index_count; ++b)
				{
					int j = index.entry[b];
					int thrA_count = thrA.entry[j].count;
					for (register int c = 1; c <= thrA_count; ++c)
					{
						register int k = thrA.entry[j].entry[c];
						if (!used.entry[k])
						{
							++new_index_count;
							index.entry[new_index_count] = k;
							used.entry[k] = true;
						}
					}
				}
				index_count = new_index_count;
			}
			++index_count;
			index.entry[index_count] = i;

			// Allocate preconditioner
			if (factorized)
			{
				int M_count_i = 0;
				for (register int b = 1; b <= index_count; ++b)
				{
					int j = index.entry[b];
					used.entry[j] = false;
					if (j <= i)
					{
						++M_count_i;
					}
				}
				M.AllocateRow(i, M_count_i);
				for (register int b = 1, c = 0; b <= index_count; ++b)
				{
					if (index.entry[b] <= i)
					{
						++c;
						M.index[i][c] = index.entry[b];
					}
				}
			}
			else
			{
				M.AllocateRow(i, index_count);
				for (register int b = 1; b <= index_count; ++b)
				{
					int j = index.entry[b];
					used.entry[j] = false;
					M.index[i][b] = index.entry[b];
				}
			}
			used.entry[i] = false;
		}
		M.SortIndexes();
		Log(1, "Pattern calculated, nnz(M) = %i", M.NonZero());
	}
	catch (Exception&)
	{
		ReThrow();
	}
}


template <typename T>
void SparseApproximateInverse(const CSRMatrix<T>& A, CSRMatrix<T>& G, CSRMatrix<T>& Gt, T threshold, int preconditioner_level, int threads) throw(Memory::Exception)
{
	try
	{
		Log(1, "SparseApproximateInverse, threshold = %e, level = %i", threshold, preconditioner_level);

		PowerPreconditionerPattern(A, G, threshold, preconditioner_level, true, threads);

		Vector<Vector<T> > thread_As(threads);
		Vector<Vector<T> > thread_L(threads);
		Vector<Vector<T> > thread_z(threads);
		Vector<Vector<T> > thread_x(threads);
		#pragma omp parallel for default(shared) schedule(guided) num_threads(threads)
		for (int s = 1; s <= G.rows; ++s)
		{
			// Solve the small SPD system, then fill the triangular superior part of M
			int N = G.count[s];
			if (N == 1)
			{
				int i = G.index[s][1];
				G.entry[s][1] = 1/A(i, i);
			}
			else if (N == 2)
			{
				int i1 = G.index[s][1];
				int i2 = G.index[s][2];
				T a = A(i1, i1);
				T b = A(i1, i2), c = A(i2, i2);
				T det = a*c - b*b;
				G.entry[s][1] = -b/det;
				G.entry[s][2] = a/det;
			}
			else if (N == 3)
			{
				int i1 = G.index[s][1];
				int i2 = G.index[s][2];
				int i3 = G.index[s][3];
				T a = A(i1, i1);
				T b = A(i2, i1), c = A(i2, i2);
				T d = A(i3, i1), e = A(i3, i2), f = A(i3, i3);
				T becd = b*e - c*d;
				T dbae = d*b - a*e;
				T acbb = a*c - b*b;
				T det = d*becd + e*dbae + f*acbb;
				G.entry[s][1] = becd/det;
				G.entry[s][2] = dbae/det;
				G.entry[s][3] = acbb/det;
			}
			else
			{
				int thread = omp_get_thread_num() + 1;
				Vector<T>& As = thread_As.entry[thread];
				Vector<T>& L = thread_L.entry[thread];
				Vector<T>& z = thread_z.entry[thread];
				Vector<T>& x = thread_x.entry[thread];
				if (x.size < N)
				{
					int triangular_size = (N*(N + 1)) >> 1;
					As.Resize(triangular_size);
					L.Resize(triangular_size);
					z.Resize(N);
					x.Resize(N);
				}

				// Fill matrix
				for (register int i = 1; i <= N; ++i)
				{
					register int ii = (i*(i - 1)) >> 1;
					int ig = G.index[s][i];
					for (register int j = 1; j <= i; ++j)
					{
						int jg = G.index[s][j];
						As.entry[ii + j] = A(ig, jg);
					}
				}
				// Cholesky factorization
				for (register int i = 1; i <= N; ++i)
				{
					register int ii = (i*(i - 1)) >> 1;
					T Lii = As.entry[ii + i];
					for (register int j = 1; j < i; ++j)
					{
						register int jj = (j*(j - 1)) >> 1;
						T Lij = As.entry[ii + j];
						for (register int k = 1; k < j; ++k)
						{
							Lij -= L.entry[ii + k]*L.entry[jj + k];
						}
						Lij /= L.entry[jj + j];
						L.entry[ii + j] = Lij;
						Lii -= Lij*Lij;
					}
					L.entry[ii + i] = sqrt(Lii);
				}
				// Forward sustitution
				for (register int i = 1; i <= N; ++i)
				{
					register int ii = (i*(i - 1)) >> 1;
					register T zi = (i == N) ? 1.0 : 0.0;
					for (register int k = 1; k < i; ++k)
					{
						zi -= L.entry[ii + k]*z.entry[k];
					}
					z.entry[i] = zi/L.entry[ii + i];
				}
				// Backward sustitution
				for (register int i = N; i; --i)
				{
					register int ii = (i*(i - 1)) >> 1;
					register T xi = z.entry[i];
					for (register int k = i + 1; k <= N; ++k)
					{
						register int kk = (k*(k - 1)) >> 1;
						xi -= L.entry[kk + i]*x.entry[k];
					}
					x.entry[i] = xi/L.entry[ii + i];
				}
				for (register int i = 1; i <= N; ++i)
				{
					G.entry[s][i] = x.entry[i];
				}
			}
		}

		// Calculate D and calculate G
		#pragma omp parallel for default(shared) schedule(guided) num_threads(threads)
		for (int i = 1; i <= G.rows; ++i)
		{
			int count = G.count[i];
			T Dii = 1/sqrt(G.entry[i][count]);
			for (register int c = 1; c <= count; ++c)
			{
				G.entry[i][c] *= Dii;
			}
		}
		Log(1, "G filled");

		Transpose(G, Gt);
		Log(1, "G' filled");
	}
	catch (Exception&)
	{
		ReThrow();
	}
}


template <typename T>
void SparseApproximateInverse(const CSRMatrix<T>& A, CSRMatrix<T>& M, CSCMatrix<T>& Mc, int preconditioner_level, int threads) throw(Memory::Exception)
{
	try
	{
		// E. Chow, Y. Saad. Approximate Inverse Preconditioners via Sparse-Sparse Iterations. SIAM Journal on Scientific Computing. Vol. 19-3, pp. 995-1023. 1998

		// Calculate alpha (Approximately. Alpha approximation good enough if A size > 10)
		T tr_A_At = 0;
		T tr_A_At_A_At = 0;
		#pragma omp parallel for default(shared) reduction(+:tr_A_At,tr_A_At_A_At) schedule(guided) num_threads(threads)
		for (int i = 1; i <= A.rows; ++i)
		{
			for (int q = 1; q <= A.count[i]; ++q)
			{
				tr_A_At += A.entry[i][q]*A.entry[i][q];

				int j = A.index[i][q];
				T sum = SparseDotMultiplication(A.index[i], A.entry[i], A.count[i], A.index[j], A.entry[j], A.count[j]);
				tr_A_At_A_At += sum*sum;
			}
		}
		T alpha = tr_A_At/tr_A_At_A_At; // alpha = tr(A*A')/tr(A*A'*(A*A')')

		// Initial preconditioner approximation M0 = alpha*A'
		Transpose(A, Mc);
		#pragma omp parallel for default(shared) num_threads(threads)
		for (int i = 1; i <= Mc.rows; ++i)
		{
			int count_i = Mc.count[i];
			register T* __restrict Mc_entry_i = Mc.entry[i];
			for (register int q = 1; q <= count_i; ++q)
			{
				Mc_entry_i[q] *= alpha;
			}
		}

		// Approximate inverse via MR iteration
		#pragma omp parallel for default(shared) schedule(guided) num_threads(threads)
		for (int j = 1; j <= Mc.columns; ++j)
		{
			int N = Mc.count[j]; // Actual size of sparse vectors

			// m = M*e_j
			CSVector<T> m(Mc.rows);
			CSVector<T> r(Mc.rows);
			m.Allocate(N);
			r.Allocate(N);
			for (register int q = 1; q <= N; ++q)
			{
				register int i = Mc.index[j][q];
				m.entry[q] = Mc.entry[j][q];
				m.index[q] = i;
				r.index[q] = i;
			}

			for (int level = 1; level <= preconditioner_level; ++level)
			{
				// Calculate residual r_j = e_j - A*m_j
				for (register int q = 1; q <= N; ++q)
				{
					register int i = r.index[q];
					r.entry[q] = -SparseDotMultiplication(A.index[i], A.entry[i], A.count[i], m.index, m.entry, N);
					if (i == j)
					{
						r.entry[q] += 1;
					}
				}

				// Calculate alpha
				T rAr = 0;
				T ArAr = 0;
				for (register int q = 1; q <= N; ++q)
				{
					register int i = r.index[q];
					T Ar_i = SparseDotMultiplication(A.index[i], A.entry[i], A.count[i], r.index, r.entry, N);
					rAr += r.entry[q]*Ar_i;
					ArAr += Ar_i*Ar_i;
				}
				T alpha = rAr/ArAr;

				// Update column
				for (register int q = 1; q <= N; ++q)
				{
					m.entry[q] += alpha*r.entry[q];
				}

				if (N == 1)
				{
					break;
				}
			}

			// Update preconditioner column
			Mc.AllocateColumn(j, N);
			for (register int q = 1; q <= N; ++q)
			{
				Mc.entry[j][q] = m.entry[q];
				Mc.index[j][q] = m.index[q];
			}
		}

		// Update preconditioner
		Convert(Mc, M);
	}
	catch (Exception&)
	{
		ReThrow();
	}
}

#endif
