// Partition.cpp
// Copyright (C) 2011 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <Basic/Debug.h>
#include <Container/List.h>
#include <Container/Set.h>
#include <Substructuring/Partition.h>

extern "C"
{
	#include <metis.h>
}


void ElementAdjacency(const Mesh& mesh, Vector<Set<int> >& adjacency) throw(Memory::Exception)
{
	try
	{
		ShapeType element_type = mesh.element_type;
		int elements_count = mesh.elements_count;
		int nodes_count = mesh.nodes_count;
		const Matrix<int>& connectivity = mesh.connectivity;

		// Generate adjacency for dual graph
		int vertex_count = (element_type == shape_triangle) ? 3 : (element_type == shape_tetrahedron) ? 4 : (element_type == shape_hexahedron) ? 8 : 4;
		adjacency.Resize(elements_count);
		{
			// Identify elements sharing a node
			Vector<List<int> > node_elements(nodes_count);
			for (int e = 1; e <= elements_count; ++e)
			{
				for (register int v = 1; v <= vertex_count; ++v)
				{
					register int n = connectivity.entry[e][v];
					node_elements.entry[n].AppendLast(e);
				}
			}

			// Create adjacency
			for (int n = 1; n <= nodes_count; ++n)
			{
				for (register ListItem<int>* __restrict l1 = node_elements.entry[n].first; l1; l1 = l1->next)
				{
					int e1 = l1->value;
					for (register ListItem<int>* __restrict l2 = l1->next; l2; l2 = l2->next)
					{
						int e2 = l2->value;
						adjacency.entry[e1].Append(e2);
						adjacency.entry[e2].Append(e1);
					}
				}
			}
		}
	}
	catch (Exception&)
	{
		ReThrow();
	}
}


void MetisPartitionDualGraph(const Vector<Set<int> >& adjacency, const int partitions_count, Vector<idxtype>& element_partition_id) throw(Memory::Exception)
{
	Assert(partitions_count >= 2);

	try
	{
		int elements_count = adjacency.size;

		// Count number of edges of the dual graph
		int total_edges = 0;
		for (register int e = 1; e <= elements_count; ++e)
		{
			total_edges += adjacency.entry[e].size;
		}

		// Generate adjacency structures for METIS
		Vector<idxtype> adjacency_start(elements_count + 1);
		Vector<idxtype> adjacency_nodes(total_edges);
		for (idxtype e = 1, ads = 1; e <= elements_count; ++e)
		{
			adjacency_start.entry[e] = ads;
			for (register SetItem<int>* __restrict s = adjacency.entry[e].first; s; s = s->next, ++ads)
			{
				adjacency_nodes.entry[ads] = (idxtype)s->value;
			}
		}
		adjacency_start.entry[elements_count + 1] = total_edges;

		// Call METIS partitioning routine for dual graph
		int wgtflag = 0; // Used to indicate if the graph is weighted
		int numflag = 1; // Used to indicate which numbering scheme is used for the adjacency structure of the graph
		int options[5] = {0, 0, 0, 0, 0};
		int edgecut; // Upon successful completion, this variable stores the number of edges that are cut by the partition.
		element_partition_id.Resize(elements_count); // Stores the partition number assigned to each element
		METIS_PartGraphKway((int*)&elements_count, adjacency_start.data, adjacency_nodes.data, (idxtype*)0, (idxtype*)0, &wgtflag, &numflag, (int*)&partitions_count, options, &edgecut, element_partition_id.data);
	}
	catch (Exception&)
	{
		ReThrow();
	}
}


void StructurePartitioning(const Mesh& mesh, const int partitions_count, Vector<Substructure>& substructures, Boundary& boundary) throw(Memory::Exception)
{
	Assert(partitions_count >= 2);

	try
	{
		int nodes_count = mesh.nodes_count;

		{
			int elements_count = mesh.elements_count;

			// Partition dual graph
			Vector<idxtype> element_partition_id;
			{
				// Generate adjacency for dual graph
				Vector<Set<int> > adjacency;
				ElementAdjacency(mesh, adjacency);

				// Partition dual graph
				MetisPartitionDualGraph(adjacency, partitions_count, element_partition_id);
			}

			// Count elements per partition
			substructures.Resize(partitions_count);
			Vector<int> partition_elements_count(partitions_count);
			partition_elements_count.Fill(0);
			for (int e = 1; e <= elements_count; ++e)
			{
				int p = element_partition_id.entry[e];
				++partition_elements_count.entry[p];
			}

			// Fill partition element_index and preliminary node_index
			for (int p = 1; p <= partitions_count; ++p)
			{
				Vector<int>& element_index = substructures.entry[p].element_index;

				element_index.Resize(partition_elements_count.entry[p]);
				for (register int e = 1, index = 0; e <= elements_count; ++e)
				{
					if (element_partition_id.entry[e] == p)
					{
						element_index.entry[++index] = e;
					}
				}
				mesh.GetElementsNodes(substructures.entry[p].element_index, true, substructures.entry[p].node_index);
			}
		}

		// Identify boundary nodes
		Vector<int> used_node(nodes_count);
		used_node.Fill(0);
		for (int p = 1; p <= partitions_count; ++p)
		{
			Vector<int>& node_index = substructures.entry[p].node_index;
			for (int i = 1; i <= node_index.size; ++i)
			{
				int n = node_index.entry[i];
				++used_node.entry[n];
			}
		}

		// Fill boundary.node_index and boundary.inverse_node_index
		int boundary_node_count = 0;
		for (int n = 1; n <= nodes_count; ++n)
		{
			if (used_node.entry[n] > 1)
			{
				++boundary_node_count;
			}
		}
		boundary.node_index.Resize(boundary_node_count);
		boundary.inverse_node_index.Resize(nodes_count);
		boundary.inverse_node_index.Fill(0x8FFFFFFF); // To identify invalid indexes
		for (int n = 1, b = 0; n <= nodes_count; ++n)
		{
			if (used_node.entry[n] > 1)
			{
				boundary.node_index.entry[++b] = n;
				boundary.inverse_node_index.entry[n] = b;
			}
		}

		// Fill boundary.element_index
		{
			List<int> boundary_elements;
			const Matrix<int>& connectivity = mesh.connectivity;
			for (int e = 1; e <= connectivity.rows; ++e)
			{
				for (int c = 1; c <= connectivity.columns; ++c)
				{
					int n = connectivity.entry[e][c];
					if (used_node.entry[n] > 1)
					{
						boundary_elements.AppendLast(e);
						break;
					}
				}
			}
			boundary.element_index.Resize(boundary_elements.size);
			int b = 0;
			for (ListItem<int>* item = boundary_elements.first; item; item = item->next)
			{
				boundary.element_index.entry[++b] = item->value;
			}
		}

		// Move boundary nodes to the end of node_index
		for (int p = 1; p <= partitions_count; ++p)
		{
			Vector<int>& node_index = substructures.entry[p].node_index;
			Vector<int> boundary_node_index(node_index.size);
			Vector<int> boundary(node_index.size);
			int boundary_count = 0;
			
			for (int i = 1; i <= node_index.size; ++i)
			{
				int n = node_index.entry[i];
				if (used_node.entry[n] > 1)
				{
					++boundary_count;
					boundary.entry[boundary_count] = i;
					boundary_node_index.entry[boundary_count] = n;
				}
			}
			if (boundary_count > 0)
			{
				for (int i = boundary.entry[1], j = boundary.entry[1] + 1, b = 2; j <= node_index.size; ++j)
				{
					if ((b <= boundary_count) && (j == boundary.entry[b]))
					{
						++b;
					}
					else
					{
						node_index.entry[i] = node_index.entry[j];
						++i;
					}
				}
				for (int b = 1; b <= boundary_count; ++b)
				{
					node_index.entry[node_index.size - boundary_count + b] = boundary_node_index.entry[b];
				}
			}
			substructures.entry[p].boundary_count = boundary_count;
		}
	}
	catch (Exception&)
	{
		ReThrow();
	}
}
